import asyncio

import pytest

from aiomisc import Service


class _TestService(Service):
    loop_on_init = None  # type: asyncio.AbstractEventLoop

    async def start(self):
        assert self.loop is asyncio.get_event_loop()


@pytest.fixture()
def service(loop: asyncio.AbstractEventLoop):
    return _TestService(loop_on_init=loop)


@pytest.fixture()
def services(service: _TestService):
    return [service]


async def test_loop_fixture(service: _TestService,
                            loop: asyncio.AbstractEventLoop):
    assert service.loop is service.loop_on_init is loop



import asyncio
from functools import wraps
from typing import Union, TypeVar


Number = Union[int, float]
T = TypeVar('T')


def timeout(value):
    def decorator(func):
        if not asyncio.iscoroutinefunction(func):
            raise TypeError("Function is not a coroutine function")

        @wraps(func)
        async def wrap(*args, **kwargs):
            loop = asyncio.get_event_loop()

            return await asyncio.wait_for(
                func(*args, **kwargs),
                timeout=value,
                loop=loop,
            )
        return wrap
    return decorator



import asyncio
import itertools
import logging.handlers
import socket
from functools import wraps, partial
from multiprocessing import cpu_count
from types import CoroutineType
from typing import Any, Iterable, Tuple

try:
    from typing import Coroutine
except ImportError:
    Coroutine = CoroutineType


try:
    import uvloop
    event_loop_policy = uvloop.EventLoopPolicy()
except ImportError:
    event_loop_policy = asyncio.DefaultEventLoopPolicy()


from .thread_pool import ThreadPoolExecutor


log = logging.getLogger(__name__)


def chunk_list(iterable: Iterable[Any], size: int):
    """
    Split list or generator by chunks with fixed maximum size.
    """

    iterable = iter(iterable)

    item = list(itertools.islice(iterable, size))
    while item:
        yield item
        item = list(itertools.islice(iterable, size))


OptionsType = Iterable[Tuple[int, int, int]]


def bind_socket(*args, address: str, port: int, options: OptionsType = (),
                reuse_addr: bool = True, reuse_port: bool = False,
                proto_name: str = 'tcp'):
    """

    :param args: which will be passed to stdlib's socket constructor (optional)
    :param address: bind address
    :param port: bind port
    :param options: Tuple of pairs which contain socket option
                    to set and the option value.
    :param reuse_addr: set socket.SO_REUSEADDR
    :param reuse_port: set socket.SO_REUSEPORT
    :param proto_name: protocol name which will be logged after binding
    :return: socket.socket
    """

    if not args:
        if ':' in address:
            args = (socket.AF_INET6, socket.SOCK_STREAM)
        else:
            args = (socket.AF_INET, socket.SOCK_STREAM)

    sock = socket.socket(*args)
    sock.setblocking(False)

    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, int(reuse_addr))
    if hasattr(socket, 'SO_REUSEPORT'):
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, int(reuse_port))
    else:
        log.warning('SO_REUSEPORT is not implemented by underlying library.')

    for level, option, value in options:
        sock.setsockopt(level, option, value)

    sock.bind((address, port))
    sock_addr = sock.getsockname()[:2]

    if sock.family == socket.AF_INET6:
        log.info('Listening %s://[%s]:%s', proto_name, *sock_addr)
    else:
        log.info('Listening %s://%s:%s', proto_name, *sock_addr)

    return sock


def create_default_event_loop(pool_size=None, policy=event_loop_policy,
                              debug=False):
    try:
        asyncio.get_event_loop().close()
    except RuntimeError:
        pass  # event loop is not created yet

    asyncio.set_event_loop_policy(policy)

    loop = asyncio.new_event_loop()
    loop.set_debug(debug)
    asyncio.set_event_loop(loop)

    pool_size = pool_size or cpu_count()
    thread_pool = ThreadPoolExecutor(pool_size, loop=loop)
    loop.set_default_executor(thread_pool)

    return loop, thread_pool


def new_event_loop(pool_size=None,
                   policy=event_loop_policy) -> asyncio.AbstractEventLoop:
    loop, thread_pool = create_default_event_loop(pool_size, policy)
    return loop


def shield(func):
    """
    Simple and useful decorator for wrap the coroutine to `asyncio.shield`.

    >>> @shield
    ... async def non_cancelable_func():
    ...     await asyncio.sleep(1)

    """

    async def awaiter(future):
        return await future

    @wraps(func)
    def wrap(*args, **kwargs):
        return wraps(func)(awaiter)(asyncio.shield(func(*args, **kwargs)))

    return wrap


class SelectResult:
    def __init__(self, length):
        self.length = length
        self.result_idx = None
        self.is_exception = None
        self.value = None

    def set_result(self, idx, value, is_exception):
        if self.result_idx is not None:
            return

        self.value = value
        self.result_idx = idx
        self.is_exception = is_exception

    def result(self):
        if self.is_exception:
            raise self.value
        return self.value

    def done(self):
        return self.result_idx is not None

    def __iter__(self):
        for i in range(self.length - 1):
            if i == self.result_idx:
                yield self.value
            yield None


def cancel_tasks(tasks: Iterable[asyncio.Task]) -> Iterable[asyncio.Task]:
    if not tasks:
        return []

    for task in tasks:
        task.cancel()

    return tasks


async def _select_waiter(idx, awaitable, result):
    try:
        ret = await awaitable
    except asyncio.CancelledError:
        raise
    except Exception as e:
        return result.set_result(idx, e, is_exception=True)

    result.set_result(idx, ret, is_exception=False)


async def select(*awaitables, return_exceptions=False, cancel=True,
                 timeout=None, wait=True, loop=None) -> SelectResult:

    loop = loop or asyncio.get_event_loop()
    result = SelectResult(len(awaitables))

    def waiter(args):
        return _select_waiter(*args, result=result)

    ensure_future = partial(asyncio.ensure_future, loop=loop)

    _, pending = await loop.create_task(
        asyncio.wait(
            map(
                loop.create_task,
                map(waiter, enumerate(map(ensure_future, awaitables)))
            ),
            timeout=timeout, loop=loop,
            return_when=asyncio.FIRST_COMPLETED,
        ),
    )

    if cancel:
        cancelling = cancel_tasks(pending)

        cancel_task = asyncio.ensure_future(
            asyncio.gather(
                *cancelling, loop=loop, return_exceptions=True
            ),
            loop=loop
        )

        if wait:
            await cancel_task

    if result.is_exception and not return_exceptions:
        result.result()

    return result



from . import io
from . import log

from .backoff import asyncbackoff
from .context import Context, get_context
from .entrypoint import entrypoint
from .iterator_wrapper import IteratorWrapper
from .periodic import PeriodicCallback
from .service import Service
from .thread_pool import threaded, threaded_iterable, ThreadPoolExecutor
from .timeout import timeout

from .utils import (
    bind_socket, chunk_list, new_event_loop, select, SelectResult, shield
)


__all__ = (
    'asyncbackoff', 'Context', 'get_context',
    'entrypoint', 'io', 'IteratorWrapper', 'log', 'PeriodicCallback',
    'Service', 'threaded', 'threaded_iterable', 'ThreadPoolExecutor',
    'timeout', 'bind_socket', 'chunk_list', 'new_event_loop', 'select',
    'SelectResult', 'shield',
)



import asyncio
import logging
import os
import typing
from asyncio.events import get_event_loop
from contextlib import suppress
from functools import wraps, partial
from socket import socket
from unittest.mock import MagicMock

import pytest

asyncio.get_event_loop = MagicMock(asyncio.get_event_loop)
asyncio.get_event_loop.side_effect = get_event_loop


try:
    import uvloop
except ImportError:
    uvloop = None


try:
    from inspect import isasyncgenfunction
except ImportError:
    def isasyncgenfunction(_):
        return False


def isasyncgenerator(func):
    if isasyncgenfunction(func):
        return True
    elif asyncio.iscoroutinefunction(func):
        return False


def pytest_configure(config):
    config.addinivalue_line("markers",
                            "forbid_get_event_loop: "
                            "fail when asyncio.get_event_loop will be called")
    config.addinivalue_line("markers",
                            "catch_loop_exceptions: "
                            "fails when unhandled loop exception "
                            "will be raised")


def pytest_addoption(parser):
    group = parser.getgroup('aiomisc plugin options')
    group.addoption('--aiomisc', action='store_true', default=True,
                    help='Use aiomisc entrypoint for run async tests')

    group.addoption('--aiomisc-debug', action='store_true', default=False,
                    help='Set debug for event loop')

    group.addoption('--aiomisc-pool-size', type=int, default=4,
                    help='Default thread pool size')

    group.addoption('--aiomisc-test-timeout', type=float, default=None,
                    help='Test timeout')


def pytest_fixture_setup(fixturedef):  # type: ignore
    func = fixturedef.func

    is_async_gen = isasyncgenerator(func)

    if is_async_gen is None:
        return

    strip_request = False
    if 'request' not in fixturedef.argnames:
        fixturedef.argnames += ('request',)
        strip_request = True

    def wrapper(*args, **kwargs):  # type: ignore
        if strip_request:
            request = kwargs.pop('request')
        else:
            request = kwargs['request']

        if 'loop' not in request.fixturenames:
            raise Exception('`loop` fixture required')

        event_loop = request.getfixturevalue('loop')

        if not is_async_gen:
            return event_loop.run_until_complete(func(*args, **kwargs))

        gen = func(*args, **kwargs)

        def finalizer():  # type: ignore
            try:
                return event_loop.run_until_complete(gen.__anext__())
            except StopAsyncIteration:  # NOQA
                pass

        request.addfinalizer(finalizer)
        return event_loop.run_until_complete(gen.__anext__())

    fixturedef.func = wrapper


@pytest.fixture
def loop_debug(request):
    return request.config.getoption('--aiomisc-debug')


@pytest.fixture
def aiomisc_test_timeout(request):
    return request.config.getoption('--aiomisc-test-timeout')


def pytest_pycollect_makeitem(collector, name, obj):  # type: ignore
    if collector.funcnamefilter(name) and asyncio.iscoroutinefunction(obj):
        return list(collector._genfunctions(name, obj))


@pytest.mark.tryfirst
def pytest_pyfunc_call(pyfuncitem):  # type: ignore
    if not asyncio.iscoroutinefunction(pyfuncitem.function):
        return

    event_loop = pyfuncitem.funcargs.get('loop', None)
    aiomisc_test_timeout = pyfuncitem.funcargs.get(
        'aiomisc_test_timeout', None
    )

    kwargs = {
        arg: pyfuncitem.funcargs[arg]
        for arg in pyfuncitem._fixtureinfo.argnames
    }

    @wraps(pyfuncitem.obj)
    async def func():
        return await asyncio.wait_for(
            pyfuncitem.obj(**kwargs),
            timeout=aiomisc_test_timeout,
            loop=event_loop
        )

    event_loop.run_until_complete(func())

    return True


@pytest.fixture
def thread_pool_size(request):
    return request.config.getoption('--aiomisc-pool-size')


@pytest.fixture
def services():
    return []


@pytest.fixture
def default_context():
    return {}


loop_autouse = os.getenv('AIOMISC_LOOP_AUTOUSE', '1') == '1'


@pytest.fixture
def thread_pool_executor():
    from aiomisc.thread_pool import ThreadPoolExecutor
    return ThreadPoolExecutor


@pytest.fixture(autouse=loop_autouse)
def event_loop_policy():
    if uvloop:
        return uvloop.EventLoopPolicy()
    return asyncio.DefaultEventLoopPolicy()


@pytest.fixture
def entrypoint_kwargs() -> dict:
    return {}


@pytest.fixture(name='loop', autouse=loop_autouse)
def _loop(event_loop_policy):
    asyncio.set_event_loop_policy(event_loop_policy)
    try:
        yield asyncio.new_event_loop()
    finally:
        asyncio.set_event_loop_policy(None)


@pytest.fixture(autouse=loop_autouse)
def loop(request, services, loop_debug, default_context, entrypoint_kwargs,
         thread_pool_size, thread_pool_executor, loop):
    from aiomisc.context import get_context
    from aiomisc.entrypoint import entrypoint

    asyncio.set_event_loop(loop)

    pool = thread_pool_executor(thread_pool_size)
    loop.set_default_executor(pool)

    get_marker = request.node.get_closest_marker
    forbid_loop_getter_marker = get_marker('forbid_get_event_loop')
    catch_unhandled_marker = get_marker('catch_loop_exceptions')

    exceptions = list()
    if catch_unhandled_marker:
        loop.set_exception_handler(lambda l, c: exceptions.append(c))

    try:
        with entrypoint(*services, pool_size=thread_pool_size,
                        debug=loop_debug, loop=loop,
                        **entrypoint_kwargs):

            ctx = get_context(loop)

            for key, value in default_context.items():
                ctx[key] = value

            if forbid_loop_getter_marker:
                asyncio.get_event_loop.side_effect = partial(
                    pytest.fail, "get_event_loop is forbidden"
                )

            yield loop

            if exceptions:
                logging.error(
                    'Unhandled exceptions found:\n\n\t%s',
                    "\n\t".join(
                        (
                            "Message: {m}\n\t"
                            "Future: {f}\n\t"
                            "Exception: {e}"
                        ).format(
                            m=e['message'],
                            f=repr(e.get('future')),
                            e=repr(e.get('exception'))
                        ) for e in exceptions)
                )
                pytest.fail("Unhandled exceptions found. See logs.")
    finally:
        with suppress(Exception):
            pool.shutdown(True)

        asyncio.get_event_loop.side_effect = get_event_loop
        loop.close()
        del loop


def get_unused_port() -> int:
    sock = socket()
    sock.bind(('', 0))
    port = sock.getsockname()[-1]
    sock.close()
    return port


@pytest.fixture
def aiomisc_unused_port_factory() -> typing.Callable[[], int]:
    return get_unused_port


@pytest.fixture
def aiomisc_unused_port() -> int:
    return get_unused_port()



import asyncio
import logging
import typing
from functools import partial

from .context import Context, get_context
from .log import basic_config, LogFormat
from .service import Service
from .utils import create_default_event_loop, event_loop_policy, shield


class Entrypoint:
    async def _start(self):
        if self.log_config:
            basic_config(
                level=self.log_level,
                log_format=self.log_format,
                buffered=True,
                loop=self.loop,
                buffer_size=self.log_buffer_size,
                flush_interval=self.log_flush_interval,
            )

        await asyncio.gather(
            *[self._start_service(svc) for svc in self.services],
            loop=self.loop
        )

    def __init__(self, *services, loop: asyncio.AbstractEventLoop = None,
                 pool_size: int = None,
                 log_level: typing.Union[int, str] = logging.INFO,
                 log_format: typing.Union[str, LogFormat] = 'color',
                 log_buffer_size: int = 1024,
                 log_flush_interval: float = 0.2,
                 log_config: bool = True,
                 policy=event_loop_policy,
                 debug: bool = False):

        """

        :param debug: set debug to event loop
        :param loop: loop
        :param services: Service instances which will be starting.
        :param pool_size: thread pool size
        :param log_level: Logging level which will be configured
        :param log_format: Logging format which will be configures
        :param log_buffer_size: Buffer size for logging
        :param log_flush_interval: interval in seconds for flushing logs
        :param log_config: if False do not configure logging
        """

        self._debug = debug
        self._loop = loop
        self._loop_owner = False
        self._thread_pool = None
        self.ctx = None
        self.log_buffer_size = log_buffer_size
        self.log_config = log_config
        self.log_flush_interval = log_flush_interval
        self.log_format = log_format
        self.log_level = log_level
        self.policy = policy
        self.pool_size = pool_size
        self.services = services
        self.shutting_down = False

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        if self._loop is None:
            self._loop, self._thread_pool = create_default_event_loop(
                pool_size=self.pool_size,
                policy=self.policy,
                debug=self._debug
            )
            self._loop_owner = True

        return self._loop

    def __del__(self):
        if self._loop and self._loop.is_closed():
            return

        if self._loop_owner:
            self._loop.close()

    def __enter__(self):
        if self.log_config:
            basic_config(
                level=self.log_level,
                log_format=self.log_format,
                loop=self.loop,
                buffered=False,
            )

        self.ctx = Context(loop=self.loop)
        self.loop.run_until_complete(self._start())
        return self.loop

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self.graceful_shutdown(exc_val)
            self.shutting_down = True
        finally:
            if not self.shutting_down:
                self.graceful_shutdown(None)

            self.ctx.close()

            if self._thread_pool:
                self._thread_pool.shutdown()

            if self._loop_owner:
                self._loop.close()

    async def _start_service(self, svc: Service):
        svc.set_loop(self.loop)

        ensure_future = partial(asyncio.ensure_future, loop=self.loop)

        start_task, ev_task = map(
            ensure_future, (svc.start(), svc.start_event.wait())
        )

        await asyncio.wait(
            (start_task, ev_task), loop=self.loop,
            return_when=asyncio.FIRST_COMPLETED,
        )

        self.loop.call_soon(svc.start_event.set)
        await ev_task

        if start_task.done():
            return await start_task

    @shield
    async def _stop_service(self, svc, exception):
        await svc.stop(exception)

    def graceful_shutdown(self, exception):
        tasks = [
            self._stop_service(svc, exception) for svc in self.services
        ]

        if not tasks:
            return

        self.loop.run_until_complete(
            asyncio.gather(*tasks, loop=self.loop, return_exceptions=True)
        )

        self.loop.run_until_complete(
            self.loop.shutdown_asyncgens()
        )


entrypoint = Entrypoint


__all__ = ('entrypoint', 'Entrypoint', 'get_context')
